"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.nativeToUi = exports.uiToNative = exports.getMintDecimals = exports.decodeAggregatorInfo = exports.SubmissionLayout = exports.AggregatorLayout = exports.getMedian = exports.createAccountInstruction = exports.zeroKey = void 0;
const web3_js_1 = require("@solana/web3.js");
const layout_1 = require("./layout");
const bn_js_1 = __importDefault(require("bn.js"));
const token_instructions_1 = require("@project-serum/serum/lib/token-instructions");
const buffer_layout_1 = require("buffer-layout");
exports.zeroKey = new web3_js_1.PublicKey(new Uint8Array(32));
async function createAccountInstruction(connection, payer, space, owner, lamports) {
    const account = new web3_js_1.Account();
    const instruction = web3_js_1.SystemProgram.createAccount({
        fromPubkey: payer,
        newAccountPubkey: account.publicKey,
        lamports: lamports ? lamports : await connection.getMinimumBalanceForRentExemption(space),
        space,
        programId: owner
    });
    return { account, instruction };
}
exports.createAccountInstruction = createAccountInstruction;
function getMedian(submissions) {
    const values = submissions
        .filter((s) => s.value != 0)
        .map((s) => s.value)
        .sort((a, b) => a - b);
    const len = values.length;
    if (len == 0) {
        return 0;
    }
    else if (len == 1) {
        return values[0];
    }
    else {
        const i = len / 2;
        return len % 2 == 0 ? (values[i] + values[i - 1]) / 2 : values[i];
    }
}
exports.getMedian = getMedian;
exports.AggregatorLayout = buffer_layout_1.struct([
    buffer_layout_1.blob(4, "submitInterval"),
    layout_1.u64("minSubmissionValue"),
    layout_1.u64("maxSubmissionValue"),
    buffer_layout_1.u8("submissionDecimals"),
    buffer_layout_1.blob(32, "description"),
    buffer_layout_1.u8("isInitialized"),
    layout_1.publicKeyLayout('owner'),
    buffer_layout_1.blob(576, "submissions")
]);
exports.SubmissionLayout = buffer_layout_1.struct([
    layout_1.u64("time"),
    layout_1.u64("value"),
    layout_1.publicKeyLayout('oracle'),
]);
function decodeAggregatorInfo(accountInfo) {
    const aggregator = exports.AggregatorLayout.decode(accountInfo.data);
    const minSubmissionValue = aggregator.minSubmissionValue;
    const maxSubmissionValue = aggregator.maxSubmissionValue;
    const submitInterval = aggregator.submitInterval;
    const description = aggregator.description.toString().trim();
    // decode oracles
    const submissions = [];
    const submissionSpace = exports.SubmissionLayout.span;
    let latestUpdateTime = new bn_js_1.default(0);
    const decimalAdj = Math.pow(10, aggregator.submissionDecimals);
    for (let i = 0; i < aggregator.submissions.length / submissionSpace; i++) {
        const submission = exports.SubmissionLayout.decode(aggregator.submissions.slice(i * submissionSpace, (i + 1) * submissionSpace));
        submission.value = submission.value / decimalAdj;
        if (!submission.oracle.equals(new web3_js_1.PublicKey(0))) {
            submissions.push(submission);
        }
        if (submission.time > latestUpdateTime) {
            latestUpdateTime = submission.time;
        }
    }
    return {
        minSubmissionValue: minSubmissionValue,
        maxSubmissionValue: maxSubmissionValue,
        submissionValue: getMedian(submissions),
        submitInterval,
        description,
        oracles: submissions.map(s => s.oracle.toString()),
        latestUpdateTime: new Date(Number(latestUpdateTime) * 1000),
    };
}
exports.decodeAggregatorInfo = decodeAggregatorInfo;
const MINT_LAYOUT = buffer_layout_1.struct([buffer_layout_1.blob(44), buffer_layout_1.u8('decimals'), buffer_layout_1.blob(37)]);
async function getMintDecimals(connection, mint) {
    if (mint.equals(token_instructions_1.WRAPPED_SOL_MINT)) {
        return 9;
    }
    const { data } = throwIfNull(await connection.getAccountInfo(mint), 'mint not found');
    const { decimals } = MINT_LAYOUT.decode(data);
    return decimals;
}
exports.getMintDecimals = getMintDecimals;
function throwIfNull(value, message = 'account not found') {
    if (value === null) {
        throw new Error(message);
    }
    return value;
}
function uiToNative(amount, decimals) {
    return new bn_js_1.default(Math.round(amount * Math.pow(10, decimals)));
}
exports.uiToNative = uiToNative;
function nativeToUi(amount, decimals) {
    return amount / Math.pow(10, decimals);
}
exports.nativeToUi = nativeToUi;
//# sourceMappingURL=utils.js.map